@@ -5,7 +5,10 @@ use pairing::{
 };
 use thiserror::Error;
 
+pub mod utils;
+pub mod ft;
 pub mod polynomial;
+pub mod worker;
 
 use polynomial::{op_tree, Polynomial};
 
@@ -55,7 +58,7 @@ impl<E: Engine> KZGWitness<E> {
 // A witness for a several elements - "w_B" in the paper. It's a single group element plus a polynomial
 #[derive(Debug, Clone, PartialEq, Eq)]
 pub struct KZGBatchWitness<E: Engine> {
-    r: Polynomial<E>,
+    r: Polynomial<E::Fr>,
     w: E::G1Affine,
 }
 
@@ -68,7 +71,7 @@ impl<E: Engine> KZGBatchWitness<E> {
         &self.w
     }
 
-    pub fn polynomial(&self) -> &Polynomial<E> {
+    pub fn polynomial(&self) -> &Polynomial<E::Fr> {
         &self.r
     }
 }
@@ -81,12 +84,14 @@ pub enum KZGError {
     PointNotOnPolynomial,
     #[error("batch opening remainder is zero!")]
     BatchOpeningZeroRemainder,
+    #[error("polynomial degree too large")]
+    PolynomialDegreeTooLarge,
 }
 
 #[derive(Debug, Clone)]
 pub struct KZGProver<'params, E: Engine> {
     parameters: &'params KZGParams<E>,
-    polynomial: Option<Polynomial<E>>,
+    polynomial: Option<Polynomial<E::Fr>>,
     commitment: Option<KZGCommitment<E>>,
 }
 
@@ -109,7 +114,7 @@ impl<'params, E: Engine> KZGProver<'params, E> {
         self.parameters
     }
 
-    pub fn commit(&mut self, polynomial: Polynomial<E>) -> KZGCommitment<E> {
+    pub fn commit(&mut self, polynomial: Polynomial<E::Fr>) -> KZGCommitment<E> {
         let mut commitment = self.parameters.g * polynomial.coeffs[0];
         for i in 0..polynomial.degree() {
             commitment += self.parameters.gs[i] * polynomial.coeffs[i + 1];
@@ -121,7 +126,7 @@ impl<'params, E: Engine> KZGProver<'params, E> {
         commitment
     }
 
-    pub fn open(&self) -> Result<Polynomial<E>, KZGError> {
+    pub fn open(&self) -> Result<Polynomial<E::Fr>, KZGError> {
         self.polynomial.clone().ok_or(KZGError::NoPolynomial)
     }
 
@@ -137,7 +142,7 @@ impl<'params, E: Engine> KZGProver<'params, E> {
         self.commitment.is_some()
     }
 
-    pub fn polynomial(&self) -> Option<&Polynomial<E>> {
+    pub fn polynomial(&self) -> Option<&Polynomial<E::Fr>> {
         self.polynomial.as_ref()
     }
 
@@ -170,7 +175,6 @@ impl<'params, E: Engine> KZGProver<'params, E> {
         }
     }
 
-    // #[cfg(any(std))]
     pub fn create_witness_batched(
         &self,
         points: &[(E::Fr, E::Fr)],
@@ -178,20 +182,22 @@ impl<'params, E: Engine> KZGProver<'params, E> {
         match self.polynomial {
             None => Err(KZGError::NoPolynomial),
             Some(ref polynomial) => {
-                let zeros: Polynomial<E> = op_tree(
+                let zeros: Polynomial<E::Fr> = op_tree(
                     points.len(),
                     &|i| {
-                        let mut coeffs = vec![-points[i].0, E::Fr::one()];
+                        let coeffs = vec![-points[i].0, E::Fr::one()];
                         Polynomial::new_from_coeffs(coeffs, 1)
                     },
                     &|a, b| a * b,
                 );
 
                 let (xs, ys): (Vec<E::Fr>, Vec<E::Fr>) = points.iter().cloned().unzip();
-                let interpolation = Polynomial::<E>::lagrange_interpolation(
-                    xs.as_slice(),
-                    ys.as_slice(),
-                );
+
+                let interpolation =
+                    Polynomial::lagrange_interpolation(xs.as_slice(), ys.as_slice());
+               
+                println!("{:#?}", interpolation);
+                println!("polynomial.degree(): {}", polynomial.degree());
 
                 let numerator = polynomial - &interpolation;
                 let (psi, rem) = numerator.long_division(&zeros);
@@ -221,7 +227,7 @@ impl<'params, E: Engine> KZGVerifier<'params, E> {
     pub fn verify_poly(
         &self,
         commitment: &KZGCommitment<E>,
-        polynomial: &Polynomial<E>,
+        polynomial: &Polynomial<E::Fr>,
     ) -> bool {
         let mut check = self.parameters.g * polynomial.coeffs[0];
         for i in 0..polynomial.degree() {
@@ -255,7 +261,7 @@ impl<'params, E: Engine> KZGVerifier<'params, E> {
         commitment: &KZGCommitment<E>,
         witness: &KZGBatchWitness<E>,
     ) -> bool {
-        let z: Polynomial<E> = op_tree(
+        let z: Polynomial<E::Fr> = op_tree(
             points.len(),
             &|i| {
                 let mut coeffs = vec![-points[i].0, E::Fr::one()];
@@ -323,6 +329,7 @@ mod tests {
     use super::*;
     use bls12_381::{Bls12, Scalar};
     use lazy_static::lazy_static;
+    use pairing::group::ff::PrimeField;
     use rand::{rngs::SmallRng, Rng, SeedableRng};
     use std::sync::Mutex;
 
@@ -341,10 +348,7 @@ mod tests {
 
     fn test_participants<'params, E: Engine>(
         params: &'params KZGParams<E>,
-    ) -> (
-        KZGProver<'params, E>,
-        KZGVerifier<'params, E>,
-    ) {
+    ) -> (KZGProver<'params, E>, KZGVerifier<'params, E>) {
         let prover = KZGProver::new(params);
         let verifier = KZGVerifier::new(params);
 
@@ -352,12 +356,9 @@ mod tests {
     }
 
     // never returns zero polynomial
-    fn random_polynomial<E: Engine>(
-        min_coeffs: usize,
-        max_coeffs: usize
-    ) -> Polynomial<E> {
+    fn random_polynomial<S: PrimeField>(min_coeffs: usize, max_coeffs: usize) -> Polynomial<S> {
         let num_coeffs = RNG_1.lock().unwrap().gen_range(min_coeffs..max_coeffs);
-        let mut coeffs = vec![E::Fr::zero(); max_coeffs];
+        let mut coeffs = vec![S::zero(); max_coeffs];
 
         for i in 0..num_coeffs {
             coeffs[i] = RNG_1.lock().unwrap().gen::<u64>().into();
@@ -371,7 +372,7 @@ mod tests {
     fn assert_verify_poly<E: Engine + Debug>(
         verifier: &KZGVerifier<E>,
         commitment: &KZGCommitment<E>,
-        polynomial: &Polynomial<E>,
+        polynomial: &Polynomial<E::Fr>,
     ) {
         assert!(
             verifier.verify_poly(&commitment, &polynomial),
@@ -384,7 +385,7 @@ mod tests {
     fn assert_verify_poly_fails<E: Engine + Debug>(
         verifier: &KZGVerifier<E>,
         commitment: &KZGCommitment<E>,
-        polynomial: &Polynomial<E>,
+        polynomial: &Polynomial<E::Fr>,
     ) {
         assert!(
             !verifier.verify_poly(&commitment, &polynomial),
@@ -509,7 +510,6 @@ mod tests {
         assert!(!verifier.verify_eval_batched(&other_points, &commitment, &witness))
     }
 
-
     #[test]
     fn test_eval_batched_all_points() {
         let params = test_setup::<Bls12, 15>();
@@ -526,5 +526,4 @@ mod tests {
         let witness = prover.create_witness_batched(points.as_slice()).unwrap();
         assert!(verifier.verify_eval_batched(points.as_slice(), &commitment, &witness));
     }
-
 }
